/*
  (c)2023 Forward Computing and Control Pty. Ltd.
  NSW Australia, www.forward.com.au
  This code is not warranted to be fit for any purpose. You may only use it at your own risk.
  This code may be freely used for both private and commercial use
  Provided this copyright is maintained.
*/
#include <Arduino.h>
#include "FS.h"
#include "SPIFFS.h"
#include "HistoryRecord.h"

#include "SafeStringReader.h"
#define LINE_LEN 265
createSafeStringReader(sfReader, LINE_LEN, "\r\n"); // allow for lines 256 long
cSF(sfToken, 80); // for reading csv values

//#define DEBUG

//struct  HistoryRecordStruct {
//  double exportWhrs;
//  double importWhrs;
//  double hotwaterWhrs;
//  double offPeakWhrs;
//  bool fullyHeated;
//  double exportWhrs_yesterday;
//  double importWhrs_yesterday;
//  double hotwaterWhrs_yesterday;
//  double offPeakWhrs_yesterday;
//  bool fullyHeated_yesterday;
//  bool midnightRestart;
//};

void printHistoryRecord(HistoryRecordStruct &historyRecord, Print* out) {
  out->print(" exportWhrs:");
  out->print(historyRecord.exportWhrs, 7);
  out->println();
  out->print(" importWhrs:");
  out->print(historyRecord.importWhrs, 7);
  out->println();
  out->print(" hotwaterWhrs:");
  out->print(historyRecord.hotwaterWhrs, 7);
  out->println();
  out->print(" offPeakWhrs:");
  out->print(historyRecord.offPeakWhrs, 7);
  out->println();
  out->print(" fullyHeated:");
  out->print(historyRecord.fullyHeated ? "true" : "false");
  out->println();
  out->print(" exportWhrs_yesterday:");
  out->print(historyRecord.exportWhrs_yesterday, 7);
  out->println();
  out->print(" importWhrs_yesterday:");
  out->print(historyRecord.importWhrs_yesterday, 7);
  out->println();
  out->print(" hotwaterWhrs_yesterday:");
  out->print(historyRecord.hotwaterWhrs_yesterday, 7);
  out->println();
  out->print(" offPeakWhrs_yesterday:");
  out->print(historyRecord.offPeakWhrs_yesterday, 7);
  out->println();
  out->print(" fullyHeated_yesterday:");
  out->print(historyRecord.fullyHeated_yesterday ? "true" : "false");
  out->print(" midnightRestart:");
  out->print(historyRecord.midnightRestart ? "true" : "false");
  out->println();
}
// fromat here must match readFromFile_HistoryRecore
static bool writeToFile_HistoryRecord(File &file, HistoryRecordStruct &historyRecord, Print* out) {
  // build up line
  cSF(sfLine, LINE_LEN);
  sfLine.print(historyRecord.exportWhrs, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.importWhrs, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.hotwaterWhrs, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.offPeakWhrs, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.fullyHeated ? "1" : "0");
  sfLine.print(',');
  sfLine.print(historyRecord.exportWhrs_yesterday, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.importWhrs_yesterday, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.hotwaterWhrs_yesterday, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.offPeakWhrs_yesterday, 7);
  sfLine.print(',');
  sfLine.print(historyRecord.fullyHeated_yesterday ? "1" : "0");
  sfLine.print(',');
  sfLine.print(historyRecord.midnightRestart ? "1" : "0");
  sfLine.println();
  if (out) {
    out->print(sfLine);
  }
  if (!file.print(sfLine)) {
    return false;
  }
  return true;
}

static bool readDoubleValue(SafeStringReader &sfReader, double &doubleValue, Print *out) {
  double value = 0.0;
  if (sfReader.nextToken(sfToken, ',')) {
#ifdef DEBUG
    if (out) {
      out->print("  Value:"); out->print(sfToken);
    }
#endif
    if (!sfToken.toDouble(value)) {
      if (out) {
        out->print("  Invalid double value '"); out->print(sfToken); out->println();
        out->println(sfReader);
      }
      return false;
    }
    // else
    doubleValue = value;
    return true;
  } //else {
  if (out) {
    out->print("  Missing Value"); out->println();
  }
  return false;
}

static bool readBoolValue(SafeStringReader &sfReader, bool &boolValue, Print *out) {
  int value = 0;
  if (sfReader.nextToken(sfToken, ',')) {
#ifdef DEBUG
    if (out) {
      out->print("  Value:"); out->print(sfToken);
    }
#endif
    if (!sfToken.toInt(value)) {
      if (out) {
        out->print("  Invalid double value '"); out->print(sfToken); out->println();
        out->println(sfReader);
      }
      return false;
    }
    // else
    boolValue = (value == 0) ? false : true;
    return true;
  } //else {
  if (out) {
    out->print("  Missing Value"); out->println();
  }
  return false;
}

static bool readFromFile_HistoryRecord(File &file, HistoryRecordStruct &historyRecord, Print *out) {
  bool rtn = true;
  if (!file.available()) {
    if (out) {
      out->print("File Empty "); out->println();
    }
    return false;
  }

  sfReader.connect(file);
  while (file.available()) {
    if (sfReader.read()) {
//#ifdef DEBUG
      if (out) {
        out->print("Read Line:"); out->print(sfReader); out->println();
      }
//#endif
      sfReader.trim();
      if (sfReader.startsWith("//")) {
        // comment
        sfReader.clear();
        continue;
      }
      if (!readDoubleValue(sfReader, historyRecord.exportWhrs, out)) {
        return false;
      }
      if (!readDoubleValue(sfReader, historyRecord.importWhrs, out)) {
        return false;
      }
      if (!readDoubleValue(sfReader, historyRecord.hotwaterWhrs, out)) {
        return false;
      }
      if (!readDoubleValue(sfReader, historyRecord.offPeakWhrs, out)) {
        return false;
      }
      if (!readBoolValue(sfReader, historyRecord.fullyHeated, out)) {
        return false;
      }
      if (!readDoubleValue(sfReader, historyRecord.exportWhrs_yesterday, out)) {
        return false;
      }
      if (!readDoubleValue(sfReader, historyRecord.importWhrs_yesterday, out)) {
        return false;
      }
      if (!readDoubleValue(sfReader, historyRecord.hotwaterWhrs_yesterday, out)) {
        return false;
      }
      if (!readDoubleValue(sfReader, historyRecord.offPeakWhrs_yesterday, out)) {
        return false;
      }
      if (!readBoolValue(sfReader, historyRecord.fullyHeated_yesterday, out)) {
        return false;
      }
      if (!readBoolValue(sfReader, historyRecord.midnightRestart, out)) {
        return false;
      }
      sfReader.clear(); // for next line
      return true;
    }
  }
  return false;
}

/* You only need to format SPIFFS. the first time you run a
   test or else use the SPIFFS. plugin to create a partition
   https://github.com/me-no-dev/arduino-esp32fs-plugin */

// SPIFFS. is the fs::FS &fs arg for these methods

#define FORMAT_SPIFFS_IF_FAILED true


static Print* debugPtr =  NULL;
bool openForRead(File & file, const char * path);

static const char HISTORY_FILE_NAME[] = "/history.csv";

void listRootDir(Print * out) {
  const char dirname[] = "/";
  int levels = 0;
  if (out) {
    out->printf("Listing directory: %s\r\n", dirname);
  }

  File root = SPIFFS.open(dirname);
  if (!root) {
    if (out) {
      out->println("- failed to open directory");
    }
    return;
  }
  if (!root.isDirectory()) {
    if (out) {
      out->println(" - not a directory");
    }
    return;
  }

  File file = root.openNextFile();
  while (file) {
    if (file.isDirectory()) {
      if (out) {
        out->print("  DIR : ");
        out->println(file.name());
      }
    } else {
      if (out) {
        out->print("  FILE: ");
        out->print(file.name());
        out->print("\tSIZE: ");
        out->println(file.size());
      }
    }
    file = root.openNextFile();
  }
}


// out is for debugging / error msgs
bool writeHistory(HistoryRecordStruct & historyRecord, Print * out) {
  bool rtn = true;
#ifdef DEBUG
  if (out) {
    out->printf("Writing file: %s\r\n", HISTORY_FILE_NAME);
  }
#endif

  File file = SPIFFS.open(HISTORY_FILE_NAME, FILE_WRITE);
  if (!file) {
    if (out) {
      out->print(" Failed to open "); out->print(HISTORY_FILE_NAME); out->println(" for writing");
    }
    rtn = false;
  }
  if (!writeToFile_HistoryRecord(file, historyRecord, NULL)) {
    if (out) {
      out->print(" Write Failed "); out->print(HISTORY_FILE_NAME); out->println();
    }
    rtn = false;
  }
#ifdef DEBUG
  if (out) {
    out->print(HISTORY_FILE_NAME); out->println(" rewritten.");
  }
#endif
  file.close();
  return rtn;
}

bool writeHistory(HistoryRecordStruct & historyRecord) {
  return writeHistory(historyRecord, NULL);
}

// pass in out for error msgs even if not debugging
bool readHistory(HistoryRecordStruct & historyRecord, Print * out) {
  int lineCounter = 0;
  cSF(sfToken, 80); // for parsing settings line
  cSF(sfLine, 200); // for input line
  listRootDir(out);
  File file;
  if (!openForRead(file, HISTORY_FILE_NAME)) {
    if (out) {
      out->print("failed to open "); out->print(HISTORY_FILE_NAME); out->println();
    }
  }
  if (!readFromFile_HistoryRecord(file, historyRecord, out)) {
  if (out) {
      out->print("failed to read "); out->print(HISTORY_FILE_NAME); out->println();
    }
    file.close();
    return false;
  }
  file.close();
  return true;
}

bool readHistory(HistoryRecordStruct & historyRecord) {
  return readHistory(historyRecord, NULL);
}

bool initFileSystem(Print * _debugPtr) {
  debugPtr = _debugPtr;
  if (!SPIFFS.begin(FORMAT_SPIFFS_IF_FAILED)) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("SPIFFS. Mount Failed");
    }
#endif
    return false;
  }
  return true;
}

bool initFileSystem() {
  return initFileSystem(NULL);
}

bool openForOverwrite(File & file, const char* path) {
  file = SPIFFS.open(path, FILE_WRITE);
  if (!file || file.isDirectory()) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- failed to open file for writing");
    }
#endif
    file.close();
    return false;
  }
  return true;
}

bool openForRead(File & file, const char * path) {
#ifdef DEBUG
  if (debugPtr) {
    debugPtr->printf("Reading file: %s\r\n", path);
  }
#endif

  file = SPIFFS.open(path);
  if (!file || file.isDirectory()) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- failed to open file for reading");
    }
#endif
    file.close();
    return false;
  }
  return true;
}

bool deleteFile(const char * path) {
#ifdef DEBUG
  if (debugPtr) {
    debugPtr->printf("Deleting file: %s\r\n", path);
  }
#endif
  if (!SPIFFS.remove(path)) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- delete failed");
    }
#endif
    return false;
  }
  return true;
}



void writeFile(const char * path, const char * message) {
#ifdef DEBUG
  if (debugPtr) {
    debugPtr->printf("Writing file: %s\r\n", path);
  }
#endif

  File file = SPIFFS.open(path, FILE_WRITE);
  if (!file) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- failed to open file for writing");
    }
#endif
    return;
  }
  if (file.print(message)) {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- file written");
    }
#endif
  } else {
#ifdef DEBUG
    if (debugPtr) {
      debugPtr->println("- write failed");
    }
#endif
  }
  file.close();
}
