// SolarHotWater Monitor (no controls)
// replaced Fire_Beetle EPS32-E with XIAO-ESP32C3  because the Fire_Beetle kept rebooting multiple times each day.

// Using ESP32 based board programmed via Arduino IDE
// follow the steps given on http://www.forward.com.au/pfod/ESP32/index.html to install ESP32 support for Arduino IDE
// NOTE: The latest versions of the ESP32 V2.0.0 - V2.0.5 Arduino board support have problems with WiFi connections
//  for the ESP32 chip (e.g. Sparkfun ESP32 Thing, tec).
// It may take 3 or more re-connections to get a stable connection after reboot or reprogramming.
// ESP32 BLE and Bluetooth are OK on V2.0.4 as is ESP32C3 WiFi (different chip).
// For ESP8266 V3.0.2 works.


// You need to modify the WLAN_SSID, WLAN_PASS settings below
// to match your network settings

/* Code generated by pfodDesignerV3 V3.0.4181
*/
/*
   (c)2023 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This generated code may be freely used for both private and commercial use
   provided this copyright is maintained.
*/

#include <esp_task_wdt.h>
#include <pfodEEPROM.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include "millisDelay.h"
#include "SafeString.h"
#include <pfodDwgs.h>
#include "Dwg_1.h"
#include "Label.h"
#include <ESPAutoWiFiConfig.h>
#include "httpGetTask.h"


int ledPin = -2; // for Sparkfun ESP32 IoT Redboard // D7; // for XIAO esp32c3 external led // -D8 for Fire_Beetle, -8; for ESP32-C3
bool highForLedOn = false; // connected from 3v3 to D7
size_t eepromOffset = 32; // allow for security if you use EEPROM.begin(size) in your code add the size here so AutoWiFi data is written after your data


//#define DEBUG
//#define CLIENT_DEBUG


// Monitor No control from this app

const unsigned MAX_HotWaterOffCount = 5; // 25sec delay before marking hotwater fully charged

int status;     // the Wifi radio's status

#include <WiFi.h>

// Download pfodESP32BufferedClient library from http://www.forward.com.au/pfod/pfodParserLibraries/index.html
// pfodESP32BufferedClient.zip contains pfodESP32BufferedClient and pfodESP32Utils
#include <pfodESP32Utils.h>
#include <pfodESP32BufferedClient.h>

// download the libraries from http://www.forward.com.au/pfod/pfodParserLibraries/index.html


#include <pfodSecurity.h>
int swap01(int); // method prototype for slider end swaps
char version[] = "V2";
pfodParser parser(version); // always have this one // create a parser with menu version string to handle the pfod messages
pfodDwgs dwgs(&parser);  // drawing support

pfodESP32BufferedClient bufferedClient;


const uint8_t MAX_CLIENTS = 4; // the default
WiFiClient clients[MAX_CLIENTS]; // hold the currently open clients
pfodParser *parserPtrs[MAX_CLIENTS]; // hold the parsers for each client
pfodESP32BufferedClient *bufferedClientPtrs[MAX_CLIENTS]; // hold the parsers for each client

const int portNo = 4989; // What TCP port to listen on for connections.

// sfTime input 00:00 to 23:59, output 12:00am to 11:59pm
// input returned unchanged on errors
void convertDayMinsToAmPm(SafeString &sfTime, unsigned long dayMins) {
  int hrs = dayMins / 60;
  int mins = dayMins - hrs * 60;

  bool isPM = false;
  if (hrs >= 12) {
    hrs -= 12;
    isPM = true;
  }
  if (hrs == 0) {
    hrs += 12; // 00:.. => 12:.. am
  }
  sfTime.clear();
  sfTime.print(hrs);
  sfTime.print(":");
  if (mins < 10) {
    sfTime.print('0');
  }
  sfTime.print(mins);
  if (isPM) {
    sfTime += "pm";
  } else {
    sfTime += "am";
  }
}


// sfTime input 00:00 to 23:59, output 12:00am to 11:59pm
// input returned unchanged on errors
void convert24hrToAmPm(SafeString &sfTime) {
  cSF(sfStr, 10);
  cSF(sfMins, 10);
  if (sfTime.capacity() < 7) {
    return; // to small
  }
  if (sfTime.length() != 5) {
    return;
  }
  int idx = sfTime.indexOf(':');
  if (idx != 2) {
    return;
  }
  sfTime.substring(sfStr, 0, idx);
  sfTime.substring(sfMins, idx);
  bool isPM = false;
  int hrs = 0;
  if (sfStr.toInt(hrs)) {
    if (hrs >= 12) {
      hrs -= 12;
      isPM = true;
    }
    if (hrs == 0) {
      hrs += 12; // 00:.. => 12:.. am
    }
    sfTime.clear();
    sfTime.print(hrs);
    sfTime += sfMins;
    if (isPM) {
      sfTime += "pm";
    } else {
      sfTime += "am";
    }
  } else {
    return; // input unchanged
  }
}


pfodAutoCmd mainDwgLoadCmd; // the main empty dwg
static bool forceMainMenuReload = true; // for testing force complete reload on every reboot
static bool forceMainDwgReload = true; // for testing force complete reload on every reboot
static unsigned long dwgRefresh = 5000;//set to 0 to disable auto refresh or 10000; to  refresh every 10sec
// when on the front screen, the mobile back button will also refresh, otherwise it goes back to the previous menu/screen

Dwg_1 _dwg_1(&parser, &dwgs); // create dwg object and add to parser processing
// Monitor No control from this app
Label hotwaterkWhrs(&parser, &dwgs); // solar + offpeak totals
Label hotwaterFull(&parser, &dwgs); // hot water either solar or boost
Label totalCost(&parser, &dwgs); //
Label yesterday(&parser, &dwgs); //
Label nameTimeDisplay(&parser, &dwgs);
cSF(sfTime, 10); // 00:00 to 23:59 => 12:00am to 12:59pm

// send main dwg that just inserts the designed dwg at the pushZero position and scaling.
bool sendMainDwg(pfodParser &parser) {
  dwgs.setOut(&parser);
  // main dwg just inserts desgined dwg
  dwgs.start(50, 80, dwgs.WHITE, false); //false means NO more to come,  background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendRefreshAndVersion(0); //need to set a version number for image refresh to work!! send the parser version to cache this image, refreshes every 10sec
  dwgs.pushZero(25, 22, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(_dwg_1).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 47, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(totalCost).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 53, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(hotwaterkWhrs).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 58, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(hotwaterFull).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 65, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(yesterday).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 74, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(nameTimeDisplay).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.end();
  return true;
}

void sendMainDwgUpdate(pfodParser &parser) { // nothing here, but insertedDwgs updated by parser.parse()
  dwgs.setOut(&parser);
  dwgs.startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  dwgs.end();
}

void hard_restart() {
  esp_task_wdt_init(1, true); // watch dog timer times out in 1sec
  esp_task_wdt_add(NULL);
  while (true); // loop here until timeout force reboot.
}



WiFiServer server(portNo);
WiFiClient client;
WiFiClient* clientPtr; // non-null if have connected client

unsigned long plot_msOffset = 0; // set by {@} response
bool clearPlot = false; // set by the {@} response code


//typedef enum {
//  readyStateUnsent      = 0,            // Client created, open not yet called
//  readyStateOpened      = 1,            // open() has been called, connected
//  readyStateHdrsRecvd   = 2,            // send() called, response headers available
//  readyStateLoading     = 3,            // receiving, partial data available
//  readyStateDone        = 4             // Request complete, all data available.
//} reqStates;



void handle_pfodParser(pfodParser &parser) {
  uint8_t cmd = parser.parse(); // parse incoming data from connection
  // parser returns non-zero when a pfod command is fully parsed and not consumed by an inserted dwg
  if (cmd != 0) { // have parsed a complete msg { to }
    if ('.' == cmd) {
      // pfodApp has connected and sent {.} , it is asking for the main menu
      if (parser.isRefresh() && !forceMainMenuReload) {
        sendMainMenuUpdate(parser); // menu is cached just send update
      } else {
        forceMainMenuReload = false;
        sendMainMenu(parser); // send back the menu designed
      }

    } else if (parser.cmdEquals(mainDwgLoadCmd)) {
      if (parser.isRefresh() && !forceMainDwgReload) { // refresh and not forceReload just send update
        sendMainDwgUpdate(parser);
      } else {
        forceMainDwgReload = false; //
        sendMainDwg(parser); // just inserts dwgs and that will update them
      }

    } else if (parser.cmdEquals('A')) { // click in A menu item that holds the main dwg
      // not used as inserted dwgs handle all the clicks
      sendMainMenuUpdate(parser); // always reply to the msg.

      // handle {@} request
    } else if ('@' == cmd) { // pfodApp requested 'current' time
      parser.print(F("{@`0}")); // return `0 as 'current' raw data milliseconds
    } else if ('!' == cmd) {
      // CloseConnection command
      closeConnection(parser.getPfodAppStream());
    } else {
      // unknown command
      parser.print(F("{}")); // always send back a pfod msg otherwise pfodApp will disconnect.
    }
  }
}

void setup() {
  // Explicitly set the ESP32 to be a WiFi-client, otherwise, it by default,
  // would try to act as both a client and an access-point and could cause
  // network-issues with your other WiFi-devices on your WiFi-network.
  //  WiFi.mode(WIFI_STA);  Fire_Beetle does not like this!!
  EEPROM.begin(512);  // only use 20bytes for pfodSecurity but reserve 512 (pfodWifiConfig uses more)

  Serial.begin(115200); // set Serial for error msgs
#ifdef DEBUG
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  Serial.println();
#endif

  // always have at least one
  parserPtrs[0] = &parser;
  bufferedClientPtrs[0] = &bufferedClient;

  // fill in the rest
  for (size_t i = 1; i < MAX_CLIENTS; i++) {
    parserPtrs[i] = new pfodParser(version);
    bufferedClientPtrs[i] = new pfodESP32BufferedClient();
  }

  // set initial display state
  hotwaterkWhrs.setTextColor(0);
  hotwaterkWhrs.setTextSize(4);
  hotwaterFull.setTextSize(3);
  hotwaterFull.setTextColor(2);
  totalCost.setTextSize(6);
  yesterday.setTextSize(2);
  nameTimeDisplay.setTextSize(4);
  nameTimeDisplay.setTextColor(12);

#ifdef DEBUG
  setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
#endif

  if (ESPAutoWiFiConfigSetup(ledPin, highForLedOn, 0)) { // check if we should start access point to configure WiFi settings
    return; // in config mode so skip rest of setup
  }


#ifdef DEBUG
  Serial.println();
  Serial.println("WiFi connected");
#endif
  wifiConnected_v = true;

  // Start the server
  server.begin();
#ifdef DEBUG
  Serial.println("Server started");
#endif

  // Print the IP address
#ifdef DEBUG
  Serial.println(WiFi.localIP());
#endif

//  setHttpGetTaskDebugOut(&Serial);
  if (!createGetHttpTask()) {
    Serial.println("createGetHttpTask failed");
  }

}

void loop() {
  //  .. other code that MUST run all the time
  if (ESPAutoWiFiConfigLoop()) {  // handle WiFi config webpages
    return;  // skip the rest of the loop until config finished
  }

  if (restart_v) {
#ifdef DEBUG
    Serial.println("Rebooting...");
#endif
    hard_restart();
  }
  bool autoSetting = true;
  if (watts3PhKW_v >= 0.0) {
    autoSetting = false; // if importing then not auto
  }
  if (updated_v) {
    updated_v = false;
    _dwg_1.updateDwg_1(watts3PhKW_v, solarHotWaterKW_v, autoSetting);
  }
  handle_pfodServerConnection();  // may update state
}

bool validClient(WiFiClient &client) {
  return (client.connected());
}

void handle_pfodServerConnection() {
  if (server.hasClient()) { // new connection
#ifdef CLIENT_DEBUG
    Serial.println("new client:");
#endif
    bool foundSlot = false;
    size_t i = 0;
    for (; i < MAX_CLIENTS; i++) {
      if (!validClient(clients[i])) { // this space if free
        foundSlot = true;
        clients[i] = server.available();
        parserPtrs[i]->connect(bufferedClientPtrs[i]->connect(&(clients[i]))); // sets new io stream to read from and write to
        break;
      }
    }
    if (!foundSlot) {
      WiFiClient newClient = server.available(); // get any new client and close it
      newClient.stop();
#ifdef CLIENT_DEBUG
      Serial.println(" NO Slots available");
#endif
    } else {
#ifdef CLIENT_DEBUG
      Serial.println(i);
#endif
    }
  }
  for (size_t i = 0; i < MAX_CLIENTS; i++) {
    if (validClient(clients[i])) {
      handle_pfodParser(*parserPtrs[i]);
    }
  }
}

void closeConnection(Stream *io) {
  if (!io) {
#ifdef CLIENT_DEBUG
    Serial.println("closeConnection: Connection stream NULL");
#endif
    return;
  }
#ifdef CLIENT_DEBUG
  Serial.print("closeConnection:");
#endif
  bool foundSlot = false;
  size_t i = 0;
  for (; i < MAX_CLIENTS; i++) {
    if ((parserPtrs[i]->getPfodAppStream() == io) && validClient(clients[i])) {
      foundSlot = true;
      break;
    }
  }
  if (foundSlot) {
#ifdef CLIENT_DEBUG
    Serial.println(i);
#endif
    // found match
    parserPtrs[i]->closeConnection(); // nulls io stream
    bufferedClientPtrs[i]->stop(); // clears client reference
    clients[i].stop();
  } else {
#ifdef CLIENT_DEBUG
    Serial.println(" Connection stream NOT found");
#endif
  }
}

void setLable() {
  float localTotalCost = totalCost_v;
  float localYesterdayTotalCost = totalCost_yesterday_v;
  float localTotalHotWaterKWhrs = totalHotWaterKWhrs_v;
  float localYesterdayTotalHotWaterKWhrs = totalHotWaterKWhrs_yesterday_v;
  bool localWasHWfullyCharged = fullyHeated_v;
  bool localYesterdayWasHWfullyCharged = fullyHeated_yesterday_v;

  cSF(sfLabel, 45);
  sfLabel = "Solar Hot Water\n";
  convertDayMinsToAmPm(sfTime, dayMins_v);
  sfLabel += sfTime;
  nameTimeDisplay.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Cost today ");
  if ((localTotalCost < 0.01) && (localTotalCost > -0.01)) {  // ~zero
    totalCost.setTextColor(0);
    localTotalCost = 0.0;
  } else if (localTotalCost > 0) {
    totalCost.setTextColor(9);
  } else {
    totalCost.setTextColor(2);
    sfLabel.print("-");
    localTotalCost = -localTotalCost;
  }
  sfLabel.print("$");
  sfLabel.print(localTotalCost, 2);
  totalCost.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Yesterday  ");
  if ((localYesterdayTotalCost < 0.01) && (localYesterdayTotalCost > -0.01)) {  // ~zero
    localYesterdayTotalCost = 0.0;
  } else if (localYesterdayTotalCost > 0) {
  } else {

    sfLabel.print("-");
    localYesterdayTotalCost = -localYesterdayTotalCost;
  }
  sfLabel.print("$");
  sfLabel.print(localYesterdayTotalCost, 2);

  sfLabel.print("  ");
  sfLabel.print(localYesterdayTotalHotWaterKWhrs, 1);
  sfLabel.print("kWhr");
  yesterday.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Hot Water ");
  sfLabel.print(localTotalHotWaterKWhrs, 1);
  sfLabel.print("kWhr");
  hotwaterkWhrs.setText(sfLabel.c_str());
  if (localWasHWfullyCharged) {
    hotwaterFull.setText("Hot Water was Fully Heated");
  } else {
    hotwaterFull.setText("");
  }
}

void sendMainMenu(pfodParser &parser) {
  setLable();
  parser.print(F("{,"));  // start a Menu screen pfod message
  parser.print("<bg w>~"); // no prompt
  parser.sendRefreshAndVersion(dwgRefresh);
  // send menu items
  parser.print(F("|+A~"));
  parser.print(mainDwgLoadCmd); // the unique load cmd
  parser.print(F("}"));  // close pfod message
}

void sendMainMenuUpdate(pfodParser &parser) {
  setLable();
  parser.print(F("{;"));  // start an Update Menu pfod message
  parser.print(F("~")); // no change to colours and size
  // send menu items
  parser.print(F("|+A")); // reload A calls for dwg z update
  parser.print(F("}"));  // close pfod message
  // ============ end of menu ===========
}


int swap01(int in) {
  return (in == 0) ? 1 : 0;
}
// ============= end generated code =========
